/* Program for testing conformance of Unicode normalization */

#include <errno.h>
#include <limits.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>

#define WINVER 0x0600
#include <windows.h>

#define LINE_LEN 1024
#define ELEM_LEN 32
#define COLUMN_CNT 5

/* Default file name of the data file */
static const char *default_name = "NormalizationTest.txt";

/* Name of the input file */
static char fname[MAX_PATH];

/* Number of current line */
static int linenum;

/* Print detailed warnings instead of a summary */
static BOOL verbose = FALSE;

static BOOL skip_pairs = FALSE;
static BOOL compose = TRUE;
static BOOL decompose = TRUE;

/* Number of total and failed tests */
static int total_tests;
static int failed_tests;

/* A Unicode surrogate pair */
struct pair { WCHAR hi, lo; };

/* Make a surrogate pair for a given codepoint */
static struct pair make_pair(LONG c)
{
    struct pair s;
    WCHAR hx, hw, lx;
    LONG hu;

    hx = (WCHAR)c;
    hu = (c >> 16) & ((1 << 5) - 1);
    hw = (WCHAR)hu - 1;
    s.hi = 0xD800 | (hw << 6) | (hx >> 10);

    lx = (WCHAR)c;
    s.lo = (WCHAR)(0xDC00 | (lx & ((1 << 10) - 1)));

    return s;
}

/* Print error message (optionally preceded by location) and exit */
static void error(BOOL pos, const char *format, ...)
{
    va_list ap;

    fputs("error: ", stderr);
    if (pos) fprintf(stderr, "%s:%d: ", fname, linenum);
    va_start(ap, format);
    vfprintf(stderr, format, ap);
    va_end(ap);
    fputc('\n', stderr);
    exit(1);
}

/* Print warning in verbose mode (optionally preceded by location) */
static void warning(BOOL pos, const char *format, ...)
{
    va_list ap;

    if (!verbose) return;

    fputs("warning: ", stderr);
    if (pos) fprintf(stderr, "%s:%d: ", fname, linenum);
    va_start(ap, format);
    vfprintf(stderr, format, ap);
    va_end(ap);
    fputc('\n', stderr);
    return;
}

/* wcscmp() replacement (Unicode-agnostic unlike lstrcmpW()) */
static int wstrcmp(const WCHAR *s1, const WCHAR *s2)
{
    for (/**/; *s1==*s2 && *s1 && *s2; s1++, s2++);
    return *s1 - *s2;
}

/* Check normalization (and print warning) */
static int check(NORM_FORM form,
    int c1, const WCHAR *s1, int c2, const WCHAR *s2)
{
    ++total_tests;

    if (wstrcmp(s1, s2) || !IsNormalizedString(form, s2, -1))
    {
        const char *formname;

        ++failed_tests;

        switch (form)
        {
        case NormalizationC:
            formname = "NFC";
            break;
        case NormalizationD:
            formname = "NFD";
            break;
        case NormalizationKC:
            formname = "NFKC";
            break;
        case NormalizationKD:
            formname = "NFKD";
            break;
        default:
            error(FALSE, "Bad normalization form");
        }

        if (verbose)
        {
            size_t k;

            fputs("warning: ", stderr);
            fprintf(stderr, "%s:%d: Bad normalization: ", fname, linenum);
            fprintf(stderr, "c%d != %s(c%d)\n", c1+1, formname, c2+1);

            for (k = 0; k < lstrlenW(s1); ++k)
            {
                fprintf(stderr, "%04X ", (int)s1[k]);
            }

            fputs("!=", stderr);

            for (k = 0; k < lstrlenW(s2); ++k)
            {
                fprintf(stderr, " %04X", (int)s2[k]);
            }

            fputc('\n', stderr);
        }

        return 1;
    }
    else
    {
        return 0;
    }
}

int main(int argc, char **argv)
{
    WCHAR norm_data[COLUMN_CNT][ELEM_LEN];
    WCHAR norm_result[COLUMN_CNT][ELEM_LEN];
    char linebuf[LINE_LEN];
    FILE *f = NULL;
    int a;

    /* Explicitly set empty file name initially */
    fname[0] = '\0';

    /* Handle command line arguments */
    for (a = 1; a < argc; ++a)
    {
        if (argv[a][0] == '-' && argv[a][1] == '\0')
        {
            /* Read from stdin instead of a file */
            f = stdin;
            strcpy(fname, "<stdin>");
        }
        else if (argv[a][0] == '-')
        {
            int b;

            /* Handle options */
            for (b = 1; b < strlen(argv[a]); ++b) switch(argv[a][b])
            {
            case 'h':
                /* Print help */
                printf("Usage:\n");
                printf("%s [-h]\n", argv[0]);
                printf("%s [-cdsv] [test_file]\n\n", argv[0]);

                printf("  -c\t" "only test composition\n");
                printf("  -d\t" "only test decomposition\n");
                printf("  -h\t" "print help\n");
                printf("  -s\t" "skip tests for codepoints above 0xFFFF\n");
                printf("  -v\t" "describe all failed tests\n\n");

                printf("test_file defaults to '%s' "
                    "in the program directory\n", default_name);
                printf("if test_file is '-', stdin is read\n");
                exit(0);
                break;

            case 'c':
                /* Skip decomposition */
                compose = TRUE;
                decompose = FALSE;
                break;

            case 'd':
                /* Skip composition */
                compose = FALSE;
                decompose = TRUE;
                break;

            case 's':
                /* Skip tests for any codepoints above 0xFFFF */
                skip_pairs = TRUE;
                break;

            case 'v':
                /* Be verbose */
                verbose = TRUE;
                break;

            default:
                error(FALSE, "Unknown option: -%c", argv[a][b]);
            }
        }
        else
        {
            if (fname[0] == '\0')
            {
                if (strlen(argv[a]) < MAX_PATH) strcpy(fname, argv[a]);
                else error(FALSE, "The provided file name is too long");
            }
            else
            {
                error(FALSE, "Only one input file is allowed");
            }
        }
    }

    /* No input file was specified, fallback to executable directory */
    if (fname[0] == '\0')
    {
        DWORD e;
        char *s;

        e = GetModuleFileNameA(NULL, fname, MAX_PATH);

        /* Make sure there is space for a concatenated string
         * no matter what the executable name is */
        if (e >= MAX_PATH - strlen(default_name))
            error(FALSE, "Path to executable is too long");

        s = strrchr(fname, '\\');

        if (s != NULL) strcpy(s + 1, default_name);
        else error(FALSE, "Could not construct path to data file");
    }

    /* Open file (if not stdin) */
    if (f == NULL) f = fopen(fname, "r");

    if (f == NULL)
        error(FALSE, "Could not open '%s': %s", fname, strerror(errno));

    /* Read file line by line */
    while (fgets(linebuf, LINE_LEN, f) != NULL)
    {
        int e, i, k, m;
        long code;
        char *begin, *end;

        ++linenum;
        begin = linebuf;

        /* Skip comments and section markers */
        if (linebuf[0] == '#' || linebuf[0] == '@') continue;

        /* Read 5 columns (separated by semicolons) */
        for (i = 0; i < COLUMN_CNT; ++i)
        {
            int j = 0;

            /* Read all codepoints in a column */
            do
            {
                code = strtol(begin, &end, 16);

                /* Skip surrogate pairs if requested */
                if (skip_pairs && code > 0xFFFF) goto skipline;

                if (begin == end || code == LONG_MAX || code == LONG_MIN)
                    error(TRUE, "Could not read number");

                begin = end;

                /* Check for potential overflow */
                if (j >= ELEM_LEN - 2) error(TRUE, "Sequence too long");

                /* Fill data structure */
                if (code <= 0xFFFF)
                {
                    norm_data[i][j++] = (WCHAR)code;
                }
                else
                {
                    struct pair s;

                    s = make_pair((LONG)code);
                    norm_data[i][j++] = s.hi;
                    norm_data[i][j++] = s.lo;
                }
            }
            while (*end == ' ');

            /* NUL-terminate the string */
            norm_data[i][j] = '\0';

            /* Move to the next column */
            if (*end == ';') ++begin;
            else error(TRUE, "Expected semicolon");

            /* Ignore any data that follows */
        }

        /* Normalization conformance tests */
        /* cX: X is the column number 1..5 */

        /* NFC: c2 == NFC(c1) == NFC(c2) == NFC(c3) */
        /* NFC: c4 == NFC(c4) == NFC(c5) */
        if (compose == TRUE) for (k = 0; k < 5; ++k)
        {
            e = NormalizeString(NormalizationC,
                norm_data[k], -1, norm_result[k], ELEM_LEN);

            if (e <= 0) warning(TRUE, "NFC normalization failed");

            m = (k > 2) ? 3 : 1;

            check(NormalizationC, m, norm_data[m], k, norm_result[k]);
        }

        /* NFD: c3 == NFD(c1) == NFD(c2) == NFD(c3) */
        /* NFD: c5 == NFD(c4) == NFD(c5) */
        if (decompose == TRUE) for (k = 0; k < 5; ++k)
        {
            e = NormalizeString(NormalizationD,
                norm_data[k], -1, norm_result[k], ELEM_LEN);

            if (e <= 0) warning(TRUE, "NFD normalization failed");

            m = (k > 2) ? 4 : 2;

            check(NormalizationD, m, norm_data[m], k, norm_result[k]);
        }

        /* NFKC: c4 == NFKC(cX) */
        if (compose == TRUE) for (k = 0; k < 5; ++k)
        {
            e = NormalizeString(NormalizationKC,
                norm_data[k], -1, norm_result[k], ELEM_LEN);

            if (e <= 0) warning(TRUE, "NFKC normalization failed");

            check(NormalizationKC, 3, norm_data[3], k, norm_result[k]);
        }

        /* NFKD: c5 == NFKD(cX) */
        if (decompose == TRUE) for (k = 0; k < 5; ++k)
        {
            e = NormalizeString(NormalizationKD,
                norm_data[k], -1, norm_result[k], ELEM_LEN);

            if (e <= 0) warning(TRUE, "NFKD normalization failed");

            check(NormalizationKD, 4, norm_data[4], k, norm_result[k]);
        }

    skipline:
        continue;
    }

    /* Check for errors */
    if (ferror(f)) error(TRUE, "I/O error");

    if (f != stdin) fclose(f);

    /* Print summary */
    if (failed_tests != 0)
        printf("%d of %d tests failed\n", failed_tests, total_tests);

    return 0;
}
